"""Event Vision Library.
=====
Provides
  1. Flexible and user-friendly interface of event data.
  2. Read/write different format and types of event data.
  3. Data loader for existing research datasets.
  4. Preprocessing.
  5. Algorithms.
  6. Visualization.
"""

from .version import __version__, git_revision as __git_version__  # type: ignore
